/* roadmap_track.c - Manage a route track: the list of street blocks.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESCRIPTION:
 *
 *   This module handles the list of street blocks that makes a route. The
 *   route is computed elsewhere (possibly by another application), loaded
 *   one sequence of blocks at a time into this module.
 *
 *   In order to speed up the rendering of a route, the track module
 *   implements a binary-tree variant of a R-Tree. The R-Tree is built
 *   the first time it is needed.
 */

#include "roadmap.h"
#include "roadmap_types.h"
#include "roadmap_track.h"


typedef struct roadmap_track_item {

   int fips;
   int line;

} RoadMapTrackItem;

static RoadMapTrackItem *RoadMapTrack = NULL;
static int RoadMapTrackCount = 0;
static int RoadMapTrackAllocated = 0;

static RoadMapTrackLine *RoadMapTrackMidpoint = NULL;
static int RoadMapTrackMidpointIndex = 0;


#define ROADMAP_TRACK_BRANCH    0
#define ROADMAP_TRACK_LEAF      1

typedef struct roadmap_track_r_node {

   RoadMapArea bounding_box[2];

   int node_type;
   int child[2];
   union {
      struct roadmap_track_r_node *child[2];
      struct roadmap_track_item   *leaf[2];
   };

} RoadMapTrackRNode;


static RoadMapTrackRNode *RoadMapTrackRTree = NULL;


static void roadmap_track_remove_subtree (RoadMapTrackRNode *parent) {

   if (parent == NULL) return;

   if (parent->node_type == ROADMAP_TRACK_BRANCH) {
      roadmap_track_remove_subtree (parent->child[0]);
      roadmap_track_remove_subtree (parent->child[1]);
   }

   free (parent);
}


static int roadmap_track_build_leaf (int left, int right) {

   int node = RoadMapTrackRNodeCount++;

   RoadMapTrackRNode[node].child[0] = left;
   RoadMapTrackRNode[node].child[1] = right;

   RoadMapTrackRNode[node].node_type - ROADMAP_TRACK_LEAF;

   roadmap_track_get_edges (RoadMapTrackRNode[node].bounding_box, left);
   roadmap_track_get_edges (RoadMapTrackRNode[node].bounding_box+1, right);

   return node;
}


static roadmap_track_build_tree (void) {

   if (RoadMapTrackRTree != NULL) {
      roadmap_track_remove_subtree (RoadMapTrackRTree);
   }

   node = roadmap_track_build_leaf (0, 1);
   ...
}


void  roadmap_track_clear (void) {
}


void  roadmap_track_add_lines (int fips, int lines[], int count) {

   int i;
   int total = RoadMapTrackCount + count;


   if (count <= 0) {
      roadmap_log (ROADMAP_FATAL, "bad line count %d", count);
   }

   if (total > RoadMapTrackAllocated) {
      RoadMapTrack = realloc (RoadMapTrack, total * sizeof(RoadMapTrackItem));
      roadmap_check_allocated(RoadMapTrack);
      RoadMapTrackAllocated = total;
   }

   for (i = RoadMapTrackCount; i < total; ++i) {
      RoadMapTrack[i].fips = fips;
      RoadMapTrack[i].line = *(lines++);
   }

   if (RoadMapTrackRTree != NULL) {
      roadmap_track_remove_subtree (RoadMapTrackRTree);
      RoadMapTrackRTree = NULL;
   }
}


void roadmap_track_behind (int fips, int line) {
}


typedef void (* RoadMapTrackAction) (int line) {
}


void  roadmap_track_visible_lines (RoadMapTrackAction action, int filter) {

   if (
}

